import logging
import json
from typing import Optional
from raga import TestSession, ModelABTestRules, FMARules, LQRules, LDTRules, EventABTestRules, Filter, OcrRules, OcrAnomalyRules, \
    DARules, FMA_LLMRules, ClassImbalanceRules, SBRules, IPDRules, LLMRules, LlmGuardrailTestRules
from raga import STRules

def model_ab_test(test_session:TestSession,
                  dataset_name: str, 
                  test_name: str, 
                  modelA: str, 
                  modelB: str,
                  type: str, 
                  rules: ModelABTestRules, 
                  aggregation_level:  Optional[list] = [],
                  gt: Optional[str] = "", 
                  filter: Optional[str] = ""):
    dataset_id = ab_test_validation(test_session, dataset_name, test_name, modelA, modelB, type, rules, gt, aggregation_level)    
    return {
            "datasetId": dataset_id,
            "experimentId": test_session.experiment_id,
            "name": test_name,
            "modelA": modelA,
            "modelB": modelB,
            "type": type,
            "rules": rules.get(),
            "aggregationLevels": aggregation_level,
            'filter':filter,
            'gt':gt,
            'test_type':'ab_test'
        }

def event_ab_test(test_session:TestSession, 
                  dataset_name: str, 
                  test_name: str, 
                  modelA: str,                  
                  modelB: str,                  
                  object_detection_modelA: str,                
                  object_detection_modelB: str,                
                  type: str, 
                  sub_type:str,
                  rules: EventABTestRules, 
                  aggregation_level:  Optional[list] = [],
                  output_type:Optional[str] = "", 
                  filter: Optional[Filter] = None):
    dataset_id = ab_event_test_validation(test_session, dataset_name, test_name, modelA, modelB, type, sub_type, rules, object_detection_modelA, object_detection_modelB, aggregation_level)    
    payload = {
            "datasetId": dataset_id,
            "experimentId": test_session.experiment_id,
            "name": test_name,
            "modelA": modelA,
            "modelB": modelB,
            "objectDetectionModelA":object_detection_modelA,
            "objectDetectionModelB":object_detection_modelB,
            "type": type,
            "subType":sub_type,
            "rules": rules.get(),
            "aggregationLevels": aggregation_level,
            'filter':"",
            'outputType':output_type,
            'test_type':'event_ab_test'
        }
    if isinstance(filter, Filter):
         payload["filter"] = filter.get()
    return payload

def ab_event_test_validation(test_session:TestSession, 
                       dataset_name: str, 
                       test_name: str, 
                       modelA: str,                        
                       modelB:str,                        
                       type: str, 
                       sub_type: str,
                       rules: ModelABTestRules,    
                       object_detection_modelA: str,
                       object_detection_modelB: str,                  
                       aggregation_level:Optional[list] = []):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"

    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
            raise ValueError(INVALID_RESPONSE)
    
    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)
    
    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(modelA, str) and modelA, f"{REQUIRED_ARG_V2.format('modelA', 'str')}"
    assert isinstance(modelB, str) and modelB, f"{REQUIRED_ARG_V2.format('modelB', 'str')}"
    assert isinstance(object_detection_modelA, str) and object_detection_modelA, f"{REQUIRED_ARG_V2.format('object_detection_modelA', 'str')}"
    assert isinstance(object_detection_modelB, str) and object_detection_modelB, f"{REQUIRED_ARG_V2.format('object_detection_modelB', 'str')}"
    assert isinstance(sub_type, str), f"{REQUIRED_ARG_V2.format('sub_type', 'str')}"
    assert isinstance(type, str), f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(rules, ModelABTestRules) and rules.get(), f"{REQUIRED_ARG_V2.format('rules', 'instance of the ModelABTestRules')}"

    if aggregation_level:
        assert isinstance(aggregation_level, list), f"{REQUIRED_ARG_V2.format('aggregation_level', 'str')}"

    return dataset_id

def ab_test_validation(test_session:TestSession, 
                       dataset_name: str, 
                       test_name: str, 
                       modelA: str, 
                       modelB: str,
                       type: str, 
                       rules: ModelABTestRules,
                       gt: Optional[str] = "", 
                       aggregation_level:Optional[list] = []):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"

    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
            raise ValueError(INVALID_RESPONSE)
    
    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)
    
    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(modelA, str) and modelA, f"{REQUIRED_ARG_V2.format('modelA', 'str')}"
    assert isinstance(modelB, str) and modelB, f"{REQUIRED_ARG_V2.format('modelB', 'str')}"
    assert isinstance(type, str), f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(rules, ModelABTestRules) and rules.get(), f"{REQUIRED_ARG_V2.format('rules', 'instance of the ModelABTestRules')}"

    if aggregation_level:
        assert isinstance(aggregation_level, list), f"{REQUIRED_ARG_V2.format('aggregation_level', 'str')}"

    if type == "labelled":
        assert isinstance(gt, str) and gt, f"{REQUIRED_ARG_V2.format('gt', 'str')}"

    if type == "unlabelled" and isinstance(gt, str) and gt:
        raise ValueError("gt is not required on unlabelled type.")
    
    return dataset_id

def failure_mode_analysis(test_session:TestSession, 
                          dataset_name:str, 
                          test_name:str, 
                          model:str, 
                          gt:str,
                          rules:FMARules,
                          output_type:str,
                          type:str,
                          clustering:Optional[dict]={},
                          aggregation_level:Optional[list]=[],
                          object_detection_model:Optional[str]="",
                          object_detection_gt:Optional[str]="",
                          embedding_col_name:Optional[str]="",
                          ):
    
    dataset_id = failure_mode_analysis_validation(test_session=test_session, dataset_name=dataset_name, test_name=test_name, model=model, gt=gt, type=type, rules=rules, output_type=output_type, aggregation_level=aggregation_level, clustering=clustering)
    response = {
            "datasetId": dataset_id,
            "experimentId": test_session.experiment_id,
            "name": test_name,
            "model": model,
            "gt": gt,
            "type": type,
            "rules": rules.get(),
            "test_type":"cluster",
            "filter":"",
            "outputType":output_type,
            "aggregationLevels":aggregation_level,
        }
    if output_type == "event_detection":
        response['objectDetectionModel'] = object_detection_model
        response['objectDetectionGT'] = object_detection_gt

    if output_type == "instance_segmentation":
        response['embeddingColName'] = embedding_col_name

    if clustering:
        response['clusterId'] = clustering
    return response

def failure_mode_analysis_validation(test_session:TestSession, dataset_name:str, test_name:str, model:str, gt:str, rules:FMARules, output_type:str, type:str, aggregation_level:Optional[list]=[], clustering:Optional[dict]=None):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
            raise ValueError(INVALID_RESPONSE)
    
    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)
    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(model, str) and model, f"{REQUIRED_ARG_V2.format('model', 'str')}"
    assert isinstance(gt, str) and gt, f"{REQUIRED_ARG_V2.format('gt', 'str')}"
    assert isinstance(type, str) and type, f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(rules, FMARules) and rules, f"{REQUIRED_ARG_V2.format('rules', 'instance of the FMARules')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"

    if output_type == "object_detection":
         if type == "embedding" and not clustering:
              raise ValueError(f"{REQUIRED_ARG_V2.format('clustering', 'clustering function')}")
         if type == "metadata":
            assert isinstance(aggregation_level, list) and aggregation_level, f"{REQUIRED_ARG_V2.format('aggregation_level', 'list')}"
    return dataset_id

def clustering(test_session:TestSession, dataset_name:str, method:str, embedding_col:str, level:str, args: dict, interpolation:bool=False, force:Optional[bool]=False):
    from raga.constants import REQUIRED_ARG_V2

    assert isinstance(method, str) and method, f"{REQUIRED_ARG_V2.format('method', 'str')}"
    assert isinstance(embedding_col, str) and embedding_col, f"{REQUIRED_ARG_V2.format('embedding_col', 'str')}"
    assert isinstance(level, str) and level, f"{REQUIRED_ARG_V2.format('level', 'str')}"
    cluster = {
        "method": method,
        "embeddingCol": embedding_col,
        "level": level,
        "args": args,
        "interpolation": interpolation
    }
    if test_session is None and dataset_name is None:
        return cluster
    else:
        dataset_id = clustering_validation(test_session, dataset_name)
        api_end_point = 'api/experiment/test/cluster'
        payload = {
            "datasetId": dataset_id,
            "forceRecreate": force,
            "clustering": cluster
        }
        from raga.constants import INVALID_RESPONSE
        from raga.utils import wait_for_status
        res_data = test_session.http_client.post(api_end_point, data=payload, headers={"Authorization": f'Bearer {test_session.token}'})
        data = res_data.get('data')
        job_id = data.get('jobId')
        if not isinstance(res_data, dict):
            raise ValueError(INVALID_RESPONSE)
        if job_id is not None:
            wait_for_status(test_session, job_id=job_id, spin=True)
        return data.get('clusterId')

def clustering_validation(test_session:TestSession, dataset_name:str):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA

    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
            raise ValueError(INVALID_RESPONSE)
    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)
    return dataset_id

def labelling_quality_test(test_session:TestSession, 
                           dataset_name:str, 
                           test_name:str, 
                           type:str, 
                           output_type: str, 
                           rules:LQRules,
                           gt:Optional[str] = "",
                           model:Optional[str] = "",
                           mistake_score_col_name: Optional[str]="",
                           train_model_column_name: Optional[str]="",
                           field_model_column_name: Optional[str]="",
                           embedding_train_col_name: Optional[str]="",
                           embedding_field_col_name: Optional[str]="",
                           embedding_col_name:Optional[str]=""):
    
    dataset_id = labelling_quality_test_validation(test_session, dataset_name, test_name, type, output_type, rules)
    payload = {
            "experimentId": test_session.experiment_id,
            "name": test_name,
            "type": type,
            "outputType": output_type,
            "rules": rules.get(),
            "filter":"",
        }
    
    if output_type == "instance_segmentation":
        payload["gt"] = gt
    
    if output_type == "object_detection":
        payload["model"] = model
        
    if train_model_column_name is not None and train_model_column_name != "" and embedding_field_col_name is not None and embedding_field_col_name != "":
        payload["trainDatasetId"]= dataset_id
        payload["fieldDatasetId"]= dataset_id
        payload["trainModelColumnName"] = train_model_column_name
        payload["fieldModelColumnName"] = field_model_column_name
        payload["embeddingTrainColName"] = embedding_train_col_name
        payload["embeddingFieldColName"] = embedding_field_col_name
        payload["test_type"] = "labelling_consistency"
    else:
        payload["datasetId"] = dataset_id
        payload["mistakeScoreColName"] = mistake_score_col_name
        payload["embeddingColName"] = embedding_col_name
        payload["test_type"] = "labelling_quality"
    
    return payload


def labelling_quality_test_validation(test_session:TestSession, dataset_name:str, test_name:str, type:str, output_type:str,  rules:LQRules):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
            raise ValueError(INVALID_RESPONSE)
    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)
    
    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(type, str) and type, f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"
    assert isinstance(rules, LQRules) and rules, f"{REQUIRED_ARG_V2.format('rules', 'str')}"
    return dataset_id

def ocr_missing_test_analysis(test_session:TestSession,
                      dataset_name:str,
                      test_name:str,
                      model:str,
                      type:str,
                      output_type: str,
                      rules:OcrRules):

    dataset_id = ocr_missing_test_analysis_validation(test_session, dataset_name, test_name, model, type, output_type, rules)
    return {
        "datasetId": dataset_id,
        "experimentId": test_session.experiment_id,
        "name": test_name,
        "model":model,
        "type": type,
        "outputType": output_type,
        "rules": rules.get(),
        "test_type":"ocr_test"
    }

def ocr_missing_test_analysis_validation(test_session:TestSession, dataset_name:str, test_name:str, model:str, type:str, output_type:str, rules:OcrRules):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
        raise ValueError(INVALID_RESPONSE)
    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)

    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(type, str) and type, f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(model, str) and model, f"{REQUIRED_ARG_V2.format('model', 'str')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"
    assert isinstance(rules, OcrRules) and rules, f"{REQUIRED_ARG_V2.format('rules', 'str')}"
    return dataset_id

def ocr_anomaly_test_analysis(test_session:TestSession,
                              dataset_name:str,
                              test_name:str,
                              model:str,
                              type:str,
                              output_type: str,
                              rules:OcrRules):

    dataset_id = ocr_anomaly_test_analysis_validation(test_session, dataset_name, test_name, model, type, output_type, rules)
    return {
        "datasetId": dataset_id,
        "experimentId": test_session.experiment_id,
        "name": test_name,
        "model":model,
        "type": type,
        "outputType": output_type,
        "rules": rules.get(),
        "test_type":"ocr_test"
    }

def ocr_anomaly_test_analysis_validation(test_session:TestSession, dataset_name:str, test_name:str, model:str, type:str, output_type:str, rules:OcrAnomalyRules):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
        raise ValueError(INVALID_RESPONSE)
    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)

    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(type, str) and type, f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(model, str) and model, f"{REQUIRED_ARG_V2.format('model', 'str')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"
    assert isinstance(rules, OcrAnomalyRules) and rules, f"{REQUIRED_ARG_V2.format('rules', 'str')}"
    return dataset_id


def active_learning(test_session:TestSession,
                    dataset_name:str,
                    test_name:str,
                    type:str,
                    output_type: str,
                    embed_col_name: str,
                    budget: int):

    dataset_id = active_learning_validation(test_session, dataset_name, test_name, type, output_type, embed_col_name, budget)
    return {
        "datasetId": dataset_id,
        "experimentId": test_session.experiment_id,
        "name": test_name,
        "type": type,
        "outputType": output_type,
        "embeddingColName": embed_col_name,
        "budget": budget,
        "test_type":"active_learning"
    }

def active_learning_validation(test_session:TestSession, dataset_name:str, test_name:str, type:str, output_type:str, embed_col_name:str, budget:int):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
        raise ValueError(INVALID_RESPONSE)
    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)

    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(type, str) and type, f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"
    assert isinstance(budget, int) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"
    assert isinstance(embed_col_name, str) and embed_col_name, f"{REQUIRED_ARG_V2.format('embed_col_name', 'str')}"
    return dataset_id

def semantic_similarity(test_session:TestSession,
                        dataset_name:str,
                        test_name:str,
                        type:str,
                        output_type: str,
                        embed_col_name: str,
                        rules: LQRules,
                        generated_embed_col_name: Optional[str] = ""):

    dataset_id = semantic_similarity_validation(test_session, dataset_name, test_name, type, output_type, embed_col_name, rules, generated_embed_col_name)
    payload =  {
        "datasetId": dataset_id,
        "experimentId": test_session.experiment_id,
        "name": test_name,
        "type": type,
        "outputType": output_type,
        "embeddingColName": embed_col_name,
        "rules": rules.get(),
    }
    if generated_embed_col_name is not None and generated_embed_col_name!="":
        payload["generatedEmbeddingColName"] = generated_embed_col_name
        payload["test_type"] = "semantic_similarity"
    else:
        payload["test_type"] = "nearest-neighbour"

    return payload

def semantic_similarity_validation(test_session:TestSession, dataset_name:str, test_name:str, type:str, output_type:str, embed_col_name:str, rules: LQRules, generated_embed_col_name: Optional[str] = ""):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
        raise ValueError(INVALID_RESPONSE)
    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)

    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(type, str) and type, f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"
    assert isinstance(embed_col_name, str) and embed_col_name, f"{REQUIRED_ARG_V2.format('embed_col_name', 'str')}"
    assert isinstance(rules, LQRules) and rules, f"{REQUIRED_ARG_V2.format('rules', 'str')}"
    return dataset_id

def nearest_duplicate(test_session:TestSession,
                      dataset_name:str,
                      test_name:str,
                      type:str,
                      output_type: str,
                      embed_col_name: str,
                      rules: LQRules):
    return semantic_similarity(test_session, dataset_name, test_name, type, output_type, embed_col_name, rules)


def data_leakage_test(test_session:TestSession,
                      dataset_name:str,
                      test_name:str,
                      type:str,
                      output_type: str,
                      rules: LQRules,
                      sub_type:Optional[str] = "",
                      train_dataset_name: Optional[str] = "",
                      train_embed_col_name: Optional[str] = "",
                      embed_col_name: Optional[str] = ""):

    dataset_id = data_leakage_test_validation(test_session, dataset_name, test_name, type, output_type)
    payload =  {
        "datasetId": dataset_id,
        "experimentId": test_session.experiment_id,
        "name": test_name,
        "type": type,
        "outputType": output_type,
        "rules": rules.get()
    }

    if train_dataset_name is not None and train_dataset_name!="" and train_embed_col_name is not None and train_embed_col_name!="" and embed_col_name is not None and embed_col_name!="":
        trainDataset_id = data_leakage_test_validation(test_session, train_dataset_name, test_name, type, output_type)
        payload["trainDatasetId"] = trainDataset_id
        payload["embeddingColName"] = embed_col_name
        payload["trainEmbeddingColName"] = train_embed_col_name
        payload["test_type"] = "data_leakage"
    else:
        payload["subType"] = sub_type
        payload["test_type"] = "data_augmentation"

    return payload

def data_leakage_test_validation(test_session:TestSession, dataset_name:str, test_name:str, type:str, output_type:str):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2
    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
        raise ValueError(INVALID_RESPONSE)
    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)

    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(type, str) and type, f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"
    return dataset_id

def data_augmentation_test(test_session:TestSession,
                      dataset_name:str,
                      test_name:str,
                      type:str,
                      sub_type:str,
                      output_type: str,
                      rules: DARules,):
    return data_leakage_test(test_session, dataset_name, test_name, type, output_type, rules, sub_type)


def label_drift_test(test_session: TestSession,
                     referenceDataset: str,
                     evalDataset: str,
                     test_name: str,
                     type: str,
                     output_type: str,
                     gt: str,
                     rules: LDTRules):

    ref_res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={referenceDataset}", headers={"Authorization": f'Bearer {test_session.token}'})
    ref_dataset_id = ref_res_data.get("data", {}).get("id")

    eval_res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={evalDataset}", headers={"Authorization": f'Bearer {test_session.token}'})
    eval_dataset_id = eval_res_data.get("data", {}).get("id")

    response = {
        "referenceDatasetId" : ref_dataset_id,
        "evalDatasetId": eval_dataset_id,
        "name": test_name,
        "gt": gt,
        "type": type,
        "test_type": "label-drift-detection",
        "outputType": output_type,
        "rules": rules.get(),
    }

    return response


def llm_guardrail_test(test_session: TestSession,
                       dataset: str,
                       test_name: str,
                       type: str,
                       test_type: str,
                       output_type: str,
                       prompt_col_name: str,
                       response_col_name: str,
                       rules: LlmGuardrailTestRules):
    dataset = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset}",
                                           headers={"Authorization": f'Bearer {test_session.token}'})
    dataset_id = dataset.get("data", {}).get("id")

    response = {
        "datasetId": dataset_id,
        "testName": test_name,
        "type": type,
        "test_type": test_type,
        "outputType": output_type,
        "promptColName": prompt_col_name,
        "responseColName": response_col_name,
        "guardrailRequestRules": rules.get(),
    }

    return response


def fma_structured_data(test_session: TestSession,
                        dataset_name: str,
                        test_name: str,
                        model: str,
                        gt: str,
                        rules: LQRules,
                        output_type: str,
                        type: str,
                        embedding: str,
                        clustering: Optional[dict] = {},
                        ):
    dataset_id = fma_structured_data_validation(test_session=test_session, dataset_name=dataset_name,
                                                test_name=test_name, model=model, gt=gt, type=type, rules=rules,
                                                output_type=output_type,
                                                clustering=clustering, embedding=embedding)
    response = {
        "datasetId": dataset_id,
        "experimentId": test_session.experiment_id,
        "name": test_name,
        "model": model,
        "gt": gt,
        "type": type,
        "rules": rules.get(),
        "test_type": "fma_sd",
        "filter": "",
        "embeddingColName":embedding,
        "outputType": output_type
    }

    if clustering:
        response['clusterId'] = clustering
    return response

def fma_structured_data_validation(test_session: TestSession, dataset_name: str, test_name: str, model: str, gt: str,
                                   rules: LQRules, output_type: str, type: str, embedding:str,
                                   clustering: Optional[dict] = None):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session,
                      TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}",
                                            headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
        raise ValueError(INVALID_RESPONSE)

    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)
    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(model, str) and model, f"{REQUIRED_ARG_V2.format('model', 'str')}"
    assert isinstance(gt, str) and gt, f"{REQUIRED_ARG_V2.format('gt', 'str')}"
    assert isinstance(type, str) and type, f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(rules,LQRules) and rules, f"{REQUIRED_ARG_V2.format('rules', 'instance of the LQRules')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"
    assert isinstance(embedding, str) and output_type, f"{REQUIRED_ARG_V2.format('embedding', 'str')}"

    return dataset_id

def scenario_imbalance(test_session: TestSession,
                        dataset_name: str,
                        test_name: str,
                        rules: SBRules,
                        output_type: str,
                        type: str,
                        embedding: Optional[str] = None,
                        clustering: Optional[dict] = {},
                        aggregationLevels: Optional[list] = []
                        ):
    dataset_id = scenario_imbalance_validation(test_session=test_session, dataset_name=dataset_name,
                                                test_name=test_name, type=type, rules=rules,
                                                output_type=output_type,
                                                clustering=clustering, embedding=embedding,
                                                aggregationLevels=aggregationLevels)
    response = {
        "datasetId": dataset_id,
        "experimentId": test_session.experiment_id,
        "name": test_name,
        "type": type,
        "rules": rules.get(),
        "test_type": "scenario_imbalance",
        "embeddingColName":embedding,
        "outputType": output_type,
        "aggregationLevels": aggregationLevels
    }

    if clustering:
        response['clusterId'] = clustering
    return response

def scenario_imbalance_validation(test_session: TestSession, dataset_name: str, test_name: str, type: str,
                                   rules: SBRules, output_type: str, embedding: Optional[str] = None,
                                   clustering: Optional[dict] = None,
                                  aggregationLevels: Optional[list] = None):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session,
                      TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}",
                                            headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
        raise ValueError(INVALID_RESPONSE)

    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)
    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(type, str) and type, f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(rules,SBRules) and rules, f"{REQUIRED_ARG_V2.format('rules', 'instance of the SBRules')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"
    return dataset_id



def failure_mode_analysis_llm(test_session:TestSession,
                              dataset_name:str,
                              test_name:str,
                              model:str,
                              gt:str,
                              rules:FMA_LLMRules,
                              type:str,
                              output_type:str,
                              prompt_col_name: str,
                              model_column: str,
                              gt_column: str,
                              embedding_col_name: str,
                              model_embedding_column: str,
                              gt_embedding_column: str,
                              clustering:Optional[dict]={},
                              ):

    dataset_id = failure_mode_analysis_llm_validation(test_session=test_session, dataset_name=dataset_name, test_name=test_name, model=model, gt=gt, rules=rules, output_type=output_type, model_column= model_column, gt_column = gt_column, model_embedding_column = model_embedding_column, gt_embedding_column =  gt_embedding_column, clustering=clustering)
    testRules = rules.get()

    metricList = {}
    for rulesDict in testRules:
        if(rulesDict["evalMetric"] == "BLEU"):
            metricList["BLEU"] = "bleu_score"
        elif(rulesDict["evalMetric"] == "ROUGE"):
            metricList["ROUGE"] = "rouge_score"
        elif(rulesDict["evalMetric"] == "METEOR"):
            metricList["METEOR"] = "meteor_score"
        elif(rulesDict["evalMetric"] == "CosineSimilarity"):
            metricList["CosineSimilarity"] = "cosine_score"
        elif(rulesDict["evalMetric"] == "user_feedback"):
            metricList["user_feedback"] = "user_feedback"

    response = {
        "datasetId": dataset_id,
        "experimentId": test_session.experiment_id,
        "name": test_name,
        "model": model,
        "gt": gt,
        "rules": rules.get(),
        "promptColName": prompt_col_name,
        "modelColName" : model_column,
        "gtColName" : gt_column,
        "embeddingColName": embedding_col_name,
        "modelEmbeddingColName" : model_embedding_column,
        "gtEmbeddingColName": gt_embedding_column,
        "test_type":"fma-llm",
        "type": type,
        "metricList": metricList,
        "outputType":output_type
    }

    if clustering:
        response['clusterId'] = clustering
    return response

def failure_mode_analysis_llm_validation(test_session:TestSession, dataset_name:str, test_name:str, model:str, gt:str, rules:FMA_LLMRules, output_type:str, model_column:str, gt_column:str, model_embedding_column:str, gt_embedding_column:str, clustering:Optional[dict]=None):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
        raise ValueError(INVALID_RESPONSE)

    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)
    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(model, str) and model, f"{REQUIRED_ARG_V2.format('model', 'str')}"
    assert isinstance(gt, str) and gt, f"{REQUIRED_ARG_V2.format('gt', 'str')}"
    assert isinstance(model_column, str) and model_column, f"{REQUIRED_ARG_V2.format('model_column', 'str')}"
    assert isinstance(gt_column, str) and gt_column, f"{REQUIRED_ARG_V2.format('gt_column', 'str')}"
    assert isinstance(model_embedding_column, str) and model_embedding_column, f"{REQUIRED_ARG_V2.format('model_embedding_column', 'str')}"
    assert isinstance(gt_embedding_column, str) and gt_embedding_column, f"{REQUIRED_ARG_V2.format('gt_embedding_column', 'str')}"
    assert isinstance(rules, FMA_LLMRules) and rules, f"{REQUIRED_ARG_V2.format('rules', 'instance of the FMARules')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"

    return dataset_id


def class_imbalance_test(test_session: TestSession,
                         dataset_name: str,
                         test_name: str,
                         type: str,
                         output_type: str,
                         annotation_column_name: str,
                         rules: ClassImbalanceRules):

    dataset_id = class_imbalance_test_validation(test_session=test_session, dataset_name=dataset_name, test_name=test_name, type=type, output_type=output_type, annotation_column_name=annotation_column_name, rules=rules)

    return {
        "datasetId": dataset_id,
        "experimentId": test_session.experiment_id,
        "name": test_name,
        "test_type": "class-imbalance",
        "type": type,
        "outputType": output_type,
        "annotationColumnName": annotation_column_name,
        "rules": rules.get(),
    }

def class_imbalance_test_validation(test_session: TestSession,
                                    dataset_name: str,
                                    test_name: str,
                                    type: str,
                                    output_type: str,
                                    annotation_column_name: str,
                                    rules: ClassImbalanceRules):

    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2
    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})

    if not isinstance(res_data, dict):
        raise ValueError(INVALID_RESPONSE)

    dataset_id = res_data.get("data", {}).get("id")
    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)

    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(type, str) and type, f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"
    assert isinstance(annotation_column_name, str) and annotation_column_name, f"{REQUIRED_ARG_V2.format('annotation_column_name', 'str')}"
    assert isinstance(rules, ClassImbalanceRules) and rules, f"{REQUIRED_ARG_V2.format('rules', 'instance of the ClassImbalanceRules')}"
    return dataset_id


def stress_test(test_session: TestSession, test_name: str,
                original_dataset: str,
                stage: str,
                output_type: str,
                gt: str,
                model: str,
                rules: STRules,
                augment_config: dict = {}):

    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2
    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(original_dataset, str) and original_dataset, f"{REQUIRED_ARG_V2.format('original_dataset', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={original_dataset}", 
                                            headers={"Authorization": f'Bearer {test_session.token}'})
    if not isinstance(res_data, dict):
        raise ValueError(INVALID_RESPONSE)

    dataset_id = res_data.get("data", {}).get("id")
    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)

    if stage not in ["augment", "stress_test"]:
        raise ValueError("stage must be augment or stress_test")

    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(stage, str) and stage, f"{REQUIRED_ARG_V2.format('stage', 'str')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"
    assert isinstance(rules, STRules) and rules, f"{REQUIRED_ARG_V2.format('rules', 'instance of the STRules')}"
    assert isinstance(gt, str) and gt, f"{REQUIRED_ARG_V2.format('gt', 'str')}"
    assert isinstance(model, str) and model, f"{REQUIRED_ARG_V2.format('model', 'str')}"

    parsed_augment_config = {}
    if stage == "augment":
        assert isinstance(augment_config, dict) and augment_config, f"{REQUIRED_ARG_V2.format('augment_config', 'dict')}"
        parsed_augment_config = stress_test_validate_augmente_config(augment_config)

    logger.debug(f'augment config : {parsed_augment_config}')

    return {
        "datasetId": dataset_id,
        "experimentId": test_session.experiment_id,
        "name": test_name,
        "test_type": "stress_test",
        "stage": stage,
        "outputType": output_type,
        "gt": gt,
        "model": model,
        "augmentConfig": parsed_augment_config,
        "rules": rules.get()
    }


def stress_test_validate_augmente_config(augment_config: dict):
    from raga.constants import REQUIRED_ARG_V2
    parsed_augment_config = {}
    for key in augment_config.keys():
        if key.startswith("scenario_"):
            scenario = augment_config[key]
            assert isinstance(scenario, dict) and scenario, f"{REQUIRED_ARG_V2.format(key, 'dict')}"

            parsed_augment_config[key] = {}
            for sub_key in scenario:
                if sub_key == "type" or sub_key == "technique" or sub_key == "default":
                    assert isinstance(scenario[sub_key], str) and scenario[sub_key], f"{REQUIRED_ARG_V2.format(sub_key, 'str')}"
                else:
                    assert isinstance(scenario[sub_key], list) and scenario[sub_key], f"{REQUIRED_ARG_V2.format(sub_key, 'list')}"
        elif "properties" == key:
            parsed_augment_config[key] = augment_config[key]
        else:
            raise KeyError(f"Invalid key in augment config : {key}")

    return json.loads(json.dumps(augment_config))


def image_property_drift(test_session: TestSession,
                         reference_dataset_name: str,
                         eval_dataset_name: str,
                         rules: IPDRules,
                         test_name: str,
                         type: str,
                         output_type: str,
                         ):
    reference_dataset_id = image_property_drift_data_validation(test_session=test_session,
                                                                dataset_name=reference_dataset_name,
                                                                test_name=test_name, rules=rules,
                                                                output_type=output_type, type=type)
    eval_dataset_id = image_property_drift_data_validation(test_session=test_session,
                                                           dataset_name=eval_dataset_name,
                                                           test_name=test_name, rules=rules,
                                                           output_type=output_type, type=type)

    response = {
        "datasetId": reference_dataset_id,
        "evalDatasetId": eval_dataset_id,
        "experimentId": test_session.experiment_id,
        "name": test_name,
        "type": type,
        "rules": rules.get(),
        "test_type": "image-property-drift",
        "outputType": output_type
    }

    return response

def image_property_drift_data_validation(test_session: TestSession, dataset_name: str, test_name: str,
                                   rules: IPDRules, output_type: str, type: str):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session,
                      TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}",
                                            headers={"Authorization": f'Bearer {test_session.token}'})


    if not isinstance(res_data, dict):
        raise ValueError(INVALID_RESPONSE)

    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)
    assert isinstance(test_name, str) and test_name, f"{REQUIRED_ARG_V2.format('test_name', 'str')}"
    assert isinstance(type, str) and type, f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(rules,IPDRules) and rules, f"{REQUIRED_ARG_V2.format('rules', 'instance of the IPDRules')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"

    return dataset_id


def llm_drift_test(test_session:TestSession,
                         train_dataset_name: str,
                         field_dataset_name: str,
                         test_name: str,
                         train_embed_col_name: str,
                         field_embed_col_name: str,
                         train_prompt_col_name: str,
                         field_prompt_col_name:  str,
                         train_response_col_name: str,
                         field_response_col_name: str,
                         model: str,
                         level: str="prompt",
                         train_gt_col_name: Optional[str] = None,
                         field_gt_col_name: Optional[str] = None,
                         train_context_col_name : Optional[str] = None,
                         field_context_col_name : Optional[str] = None,
                         rules:list=[],
                         aggregation_level:  Optional[list] = [],
                         filter: Optional[str] = ""):


    train_res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={train_dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})
    train_dataset_id = train_res_data.get("data", {}).get("id")
    field_res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={field_dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})
    field_dataset_id = field_res_data.get("data", {}).get("id")

    response = {
        "datasetId": field_dataset_id,
        "experimentId": test_session.experiment_id,
        "name": test_name,
        "aggregationLevels": aggregation_level,
        "filter":filter,
        "trainDatasetId":train_dataset_id,
        "level": level,
        "outputType":"llm-drift",
        "trainPromptColName" : train_prompt_col_name,
        "trainResponseColName" : train_response_col_name,
        "trainGtColName" : train_gt_col_name,
        "fieldPromptColName" : field_prompt_col_name,
        "fieldResponseColName" : field_response_col_name,
        "fieldGtColName" : field_gt_col_name,
        "trainEmbedColName": train_embed_col_name,
        "fieldEmbedColName": field_embed_col_name,
        "trainContextColName": train_context_col_name,
        "fieldContextColName": field_context_col_name,
        "model": model,
        "rules": rules.get(),
        'test_type':'llm-drift'
    }
    return response

def llm_performance_test(test_session:TestSession,
                              dataset_name:str,
                              model:str,
                              prompt_col_name: str,
                              context_col_name:str,
                              response_col_name:str,
                              type: str,
                              output_type: str,
                              open_ai_api_key: str,
                              rules:LLMRules,
                              test_name: str,
                              prompt_embedding_col: Optional[str] =None,
                              response_embedding_col: Optional[str] =None,
                              gt_embedding_col: Optional[str] =None,
                              gt_col_name: Optional[str] =None
                              ):

    dataset_id = llm_performance_validation(test_session=test_session,  dataset_name=dataset_name, model= model, prompt_col_name= prompt_col_name, context_col_name= context_col_name, response_col_name = response_col_name, type= type, output_type = output_type, open_ai_api_key = open_ai_api_key, rules = rules)

    response = {
        "type": type,
        "datasetId": dataset_id,
        "outputType": output_type,
        "model": model,
        "name": test_name,
        "responseColumnName": response_col_name,
        "groundTruthColumnName": gt_col_name,
        "contextColumnName": context_col_name,
        "promptColumnName": prompt_col_name,
        "openAIAPIKey": open_ai_api_key,
        "rules": rules.get(),
        "promptEmbeddingColName": prompt_embedding_col,
        "gtEmbeddingColumnName": gt_embedding_col,
        "responseEmbeddingColumn": response_embedding_col,
        "experimentId": test_session.experiment_id,
        "test_type":"llm-performance"
    }

    return response


def llm_performance_validation(test_session:TestSession, dataset_name:str, model:str, prompt_col_name: str, context_col_name:str, response_col_name:str, type: str, output_type: str, open_ai_api_key: str, rules:LLMRules):
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2

    assert isinstance(test_session,
                      TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}",
                                            headers={"Authorization": f'Bearer {test_session.token}'})


    if not isinstance(res_data, dict):
        raise ValueError(INVALID_RESPONSE)

    dataset_id = res_data.get("data", {}).get("id")

    if not dataset_id:
        raise KeyError(INVALID_RESPONSE_DATA)
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    assert isinstance(prompt_col_name, str) and prompt_col_name, f"{REQUIRED_ARG_V2.format('prompt_col_name', 'str')}"
    assert isinstance(context_col_name, str) and context_col_name, f"{REQUIRED_ARG_V2.format('context_col_name', 'str')}"
    assert isinstance(response_col_name, str) and response_col_name, f"{REQUIRED_ARG_V2.format('response_col_name', 'str')}"
    assert isinstance(open_ai_api_key, str) and open_ai_api_key, f"{REQUIRED_ARG_V2.format('open_ai_api_key', 'str')}"
    assert isinstance(model, str) and model, f"{REQUIRED_ARG_V2.format('model', 'str')}"
    assert isinstance(type, str) and type, f"{REQUIRED_ARG_V2.format('type', 'str')}"
    assert isinstance(rules, LLMRules) and rules, f"{REQUIRED_ARG_V2.format('rules', 'instance of the LLMRules')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"

    return dataset_id



import datetime
import json
import logging
import sys
import time


import requests
from raga.constants import AWS_RAGA_SECRET_KEY, AWS_RAGA_ACCESS_KEY, AWS_RAGA_ROLE_ARN
from raga.validators.test_session_validation import TestSessionValidator
from raga.utils import HTTPClient
from raga.utils import read_raga_config, get_config_value
from raga import spinner

logger = logging.getLogger(__name__)


class TestSession():
    MAX_RETRIES = 6
    RETRY_DELAY = 1
    ACCESS_KEY = "raga_access_key_id"
    SECRET_KEY = "raga_secret_access_key"



    def __init__(self, 
                 project_name: str, 
                 run_name: str = None,
                 u_test=False,
                 host=None,
                 access_key=None,
                 secret_key=None,
                 aws_raga_secret_key=None,
                 aws_raga_access_key=None,
                 aws_raga_role_arn=None,
                 profile=None,
                 inline_raga_config=False):
        spinner.start()
        
        config_data = read_raga_config(profile=profile, inline_raga_config=inline_raga_config)
        self.api_host = host if host else get_config_value(config_data, 'api_host')
        self.raga_access_key_id = access_key if access_key else get_config_value(config_data, self.ACCESS_KEY)
        self.raga_secret_access_key = secret_key if secret_key else get_config_value(config_data, self.SECRET_KEY)
        self.aws_raga_secret = aws_raga_secret_key if aws_raga_secret_key else get_config_value(config_data, AWS_RAGA_SECRET_KEY)
        self.aws_raga_access = aws_raga_access_key if aws_raga_access_key else get_config_value(config_data, AWS_RAGA_ACCESS_KEY)
        self.aws_raga_arn = aws_raga_role_arn if aws_raga_role_arn else get_config_value(config_data, AWS_RAGA_ROLE_ARN)

        self.project_name = TestSessionValidator.validate_project_name(project_name)
        if run_name is not None:
            self.run_name = TestSessionValidator.validate_run_name(run_name)
        else:
            self.run_name = run_name
        self.http_client = HTTPClient(self.api_host)
        self.test_list = []
        self.added = False

        self.token = None
        self.project_id = None
        self.experiment_id = None
        if not u_test:
            self.initialize()
        spinner.stop()


    def initialize(self):
        retries = 0
        while retries < self.MAX_RETRIES:
            try:
                self.token = self.create_token()
                self.project_id = self.get_project_id()
                break  # Exit the loop if initialization succeeds
            except requests.exceptions.RequestException as exception:
                print(f"Network error occurred: {str(exception)}")
                retries += 1
                if retries < self.MAX_RETRIES:
                    print(f"Retrying in {self.RETRY_DELAY} second(s)...")
                    time.sleep(self.RETRY_DELAY)

    def add(self, payload):
        if not isinstance(payload, dict) or not payload:
            raise ValueError("payload must be a non-empty dictionary.")
        elif payload.get("test_type") == "ab_test":
            if payload.get("type") == "unlabelled":
                payload["api_end_point"] =  "api/experiment/test/unlabelled"
            if payload.get("type") == "labelled":
                payload["api_end_point"] = "api/experiment/test/labelled"
                
        elif payload.get("test_type") == "event_ab_test":
            payload["api_end_point"] = "api/experiment/test/ab-event/v2"
            
        elif payload.get("test_type") == "drift_test":
            payload["api_end_point"] = "api/experiment/test/drift/v2"

        elif payload.get("type") == "label_drift":
            payload["api_end_point"] = "api/experiment/test/label-drift-detection"

        elif payload.get("test_type") == "guardrail":
            payload["api_end_point"] = "api/experiment/test/guardrail"
            
        elif payload.get("test_type") == "cluster":
            payload["api_end_point"] =  "api/experiment/test/fma/v2"
            if payload.get("outputType") == "event_detection":
                payload["api_end_point"] =  "api/experiment/test/fma-event/v2"
            if payload.get("outputType") == "instance_segmentation":
                payload["api_end_point"] = "api/experiment/test/fma/instance-segmentation"


        elif payload.get("test_type") == "ocr_test":
            if payload.get("outputType")  == "missing_value":
                payload["api_end_point"] =  "api/experiment/test/ocr"
            elif payload.get("outputType")  == "anomaly_detection":
                payload["api_end_point"] =  "api/experiment/test/ocr/outlier-detection/v2"
                
        elif payload.get("test_type") == "labelling_quality" and payload.get("outputType") == "instance_segmentation":
            payload["api_end_point"] = "api/experiment/test/instance-segmentation/labelling-quality"
                
        elif payload.get("test_type") == "labelling_quality":
            payload["api_end_point"] =  "api/experiment/test/labelling-quality/v2"
        elif payload.get("test_type") == "labelling_consistency":
            payload["api_end_point"] =  "api/experiment/test/labelling-consistency/v2"
        elif payload.get("test_type") == "active_learning":
            payload["api_end_point"] =  "api/experiment/test/active-learning"
        elif payload.get("test_type") == "semantic_similarity":
            payload["api_end_point"] =  "api/experiment/test/semantic-similarity/v2"
        elif payload.get("test_type") == "nearest-neighbour":
            payload["api_end_point"] =  "api/experiment/test/near-duplicate/v2"
        elif payload.get("test_type") == "data_leakage":
            payload["api_end_point"] =  "api/experiment/test/data-leakage/v2"
        elif payload.get("test_type") == "data_augmentation":
            payload["api_end_point"] =  "api/experiment/test/data-augmentation"
        elif payload.get("test_type") == "fma-llm":
            payload["api_end_point"] =  "api/experiment/test/fma-llm/v2"
        elif  payload.get("test_type") == "fma_sd":
            payload["api_end_point"] =  "api/experiment/test/fma/structured-data"
        elif payload.get("test_type") == "class-imbalance":
            payload["api_end_point"] = "api/experiment/test/class-imbalance"
        elif payload.get("test_type") == "scenario_imbalance":
            payload["api_end_point"] = "api/experiment/test/scenario-imbalance"
        elif payload.get("test_type") == "stress_test":
            payload["api_end_point"] = "api/experiment/test/stress-test"
        elif payload.get("test_type") == "image-property-drift":
            payload["api_end_point"] = "api/experiment/test/image-property-drift"
        elif payload.get("test_type") == "llm-drift":
            payload["api_end_point"] = "api/experiment/test/llm-drift"
        elif payload.get("test_type") == "llm-performance":
            payload["api_end_point"] = "api/experiment/llm/performance-test"
        self.test_list.append(payload)
        self.added = True

    def  run(self):
        from raga.constants import INVALID_RESPONSE
        if self.run_name is not None:
            self.experiment_id = self.create_experiment()
        if self.experiment_id is None:
            raise ValueError("No run name passed in test session")
        # Check if already added
        if not self.added:
            raise ValueError("add() is not called. Call add() before run().")
        if not len(self.test_list):
            raise ValueError("Test not found.")
        retries = 0
        while retries < self.MAX_RETRIES:
            try:
                for test_payload in self.test_list:
                    api_end_point = test_payload.get("api_end_point")
                    test_payload['experimentId'] = self.experiment_id
                    test_payload.pop("api_end_point")
                    test_payload.pop("test_type")
                    res_data = self.http_client.post(api_end_point, data=test_payload, headers={"Authorization": f'Bearer {self.token}'})
                    if not isinstance(res_data, dict):
                        raise ValueError(INVALID_RESPONSE)
                    logger.debug(res_data.get('data', ''))
                    self.test_list = []
                spinner.succeed("Succeed!")
                break  # Exit the loop if initialization succeeds
            except requests.exceptions.RequestException as exception:
                print(f"Network error occurred: {str(exception)}")
                retries += 1
                if retries < self.MAX_RETRIES:
                    print(f"Retrying in {self.RETRY_DELAY} second(s)...")
                    time.sleep(self.RETRY_DELAY)
                else:
                    spinner.fail("Fail!")
                self.test_list = []
                spinner.stop()
    def create_token(self):
        """
        Creates an authentication token by sending a request to the Raga API.

        Returns:
            str: The authentication token.

        Raises:
            KeyError: If the response data does not contain a valid token.
        """
        from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA
        res_data = self.http_client.post(
            "api/token",
            {"accessKey": self.raga_access_key_id, "secretKey": self.raga_secret_access_key},
        )
        if not isinstance(res_data, dict):
            raise ValueError(INVALID_RESPONSE)
        token = res_data.get("data", {}).get("token")
        if not token:
            print("access key and secret key expired, please create test session with new raga access and secret key")
            sys.exit(1)
            # raise KeyError(INVALID_RESPONSE_DATA)
        return token


    def get_project_id(self):
        """
        Get project id by sending a request to the Raga API.

        Returns:
            str: The ID of the project.

        Raises:
            KeyError: If the response data does not contain a valid project ID.
            ValueError: If the response data is not in the expected format.
        """
        from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA
        res_data = self.http_client.get(
            "api/project",
            params={"name": self.project_name},
            headers={"Authorization": f'Bearer {self.token}'},
        )

        if not isinstance(res_data, dict):
            raise ValueError(INVALID_RESPONSE)

        project_id = res_data.get("data", {}).get("id")

        if not project_id:
            raise KeyError(INVALID_RESPONSE_DATA)
        return project_id



    def create_experiment(self):
        """
        Creates an experiment by sending a request to the Raga API.

        Returns:
            str: The ID of the created experiment.

        Raises:
            KeyError: If the response data does not contain a valid experiment ID.
            ValueError: If the response data is not in the expected format.
        """
        from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA
        res_data = self.http_client.post(
            "api/experiment",
            {"name": self.run_name, "projectId": self.project_id},
            {"Authorization": f'Bearer {self.token}'},
        )

        if not isinstance(res_data, dict):
            raise ValueError(INVALID_RESPONSE)

        experiment_id = res_data.get("data", {}).get("id")

        if experiment_id is None:
            raise KeyError(INVALID_RESPONSE_DATA)
        return experiment_id



import ast
import json
import os
import sys
import time

import requests
from typing import Optional
import pandas as pd
import logging
import zipfile
import boto3


from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA


from raga.exception import RagaException
from raga.raga_schema import RagaSchema, ImageDetectionObject, ObjectDetection, PredictionSchemaElement, \
    ImageUriSchemaElement, AttributeSchemaElement, InferenceSchemaElement, StringElement, TimeStampElement, \
    TIFFSchemaElement, ImageEmbeddingSchemaElement, ImageEmbedding, Embedding, MistakeScoreSchemaElement, MistakeScore
from raga.validators.dataset_validations import DatasetValidator
from raga.dataset_creds import DatasetCreds
from raga import TestSession, spinner, Filter

from raga.utils import (get_file_name, 
                        delete_files, 
                        upload_file, 
                        create_csv_and_zip_from_data_frame, 
                        data_frame_extractor, 
                        make_arg, 
                        on_upload_success, 
                        on_upload_failed, 
                        check_key_value_existence,
                        wait_for_status,
                        PollingException)


logger = logging.getLogger(__name__)


class DatasetException(RagaException):
    pass


class DATASET_TYPE:
    IMAGE = "image"
    VIDEO = "video"
    ROI = "roi"
    EVENT = "event"
    PROMPT = "prompt"


class Dataset:
    MAX_RETRIES = 6
    RETRY_DELAY = 1

    def __init__(
        self,
        test_session:TestSession,
        name: str,
        type:Optional[str]= None,
        data: (pd.DataFrame, str) = None,
        schema: Optional[RagaSchema] = None,
        creds: Optional[DatasetCreds] = None,
        parent_dataset: Optional[str] = "",
        format: Optional[str] = None,
        model_name: Optional[str] = None,
        inference_col_name: Optional[str] = None,
        embedding_col_name: Optional[str] = None,
        filter: Optional[Filter] = None,
        model_inference_col_name: Optional[str] = None,
        event_inference_col_name: Optional[str] = None,
        u_test: bool = False,
        temp:bool = False
    ):
        spinner.start()
        self.test_session = test_session
        self.name = DatasetValidator.validate_name(name)
        self.creds = DatasetValidator.validate_creds(creds)
        self.type = type
        self.parent_dataset = parent_dataset
        self.filter = filter
        self.model_inference_col_name=model_inference_col_name
        self.event_inference_col_name = event_inference_col_name
        self.csv_file = f"experiment_{time.time_ns()}_{self.name}.csv"
        self.zip_file = f"experiment_{time.time_ns()}_{self.name}.zip"
        self.dataset_id = None
        self.temp = temp
        if not u_test and not model_inference_col_name:
            self.dataset_id = self.create_dataset()
            if self.creds and self.dataset_id:
                self.create_dataset_creds()

        self.dataset_file_id = None
        self.data_set_top_five_rows = {}
        self.raga_dataset = data
        self.raga_extracted_dataset = None
        self.raga_schema = schema
        self.dataset_schema_columns = None
        self.format = format
        self.model_name = model_name
        self.inference_col_name = inference_col_name
        self.embedding_col_name = embedding_col_name
        self.parent_dataset_id = self.parent_dataset_validation()
        self.init()

    def init(self):
        from raga import EMPTY_DATA_FRAME, INVALID_DATA_ARG, INVALID_SCHEMA, EMPTY_TEMP_DATA_FRAME
        from raga.utils.dataset_util import ds_temp_get_set
        if self.temp:
            self.raga_dataset = ds_temp_get_set(self)
            if self.raga_dataset.empty:
                raise DatasetException(EMPTY_TEMP_DATA_FRAME)
            
        if isinstance(self.raga_dataset, str) and not self.model_inference_col_name:
            pass
        elif isinstance(self.raga_dataset, pd.DataFrame) and not self.model_inference_col_name:
            if self.raga_dataset.empty:
                raise DatasetException(EMPTY_DATA_FRAME)
                
            if self.raga_schema is None:
                raise DatasetException(INVALID_SCHEMA)
            
            if not isinstance(self.raga_schema, RagaSchema):
                raise DatasetException(INVALID_SCHEMA)
            
            self.dataset_schema_columns = self.raga_schema.columns
            self.raga_extracted_dataset = data_frame_extractor(self.raga_dataset)
        elif self.event_inference_col_name and self.model_inference_col_name:
            res_data = self.test_session.http_client.get(f"api/dataset?projectId={self.test_session.project_id}&name={self.name}", headers={"Authorization": f'Bearer {self.test_session.token}'})

            if not isinstance(res_data, dict):
                    raise ValueError(INVALID_RESPONSE)
            
            self.dataset_id = res_data.get("data", {}).get("id")

            if not self.dataset_id:
                raise KeyError(INVALID_RESPONSE_DATA)  
        else:
            raise DatasetException(INVALID_DATA_ARG)

    def load(self,  schema: Optional[RagaSchema] = None, org=None):
        self.type = DatasetValidator.validate_type(self.type)
        self.raga_schema = schema if schema else self.raga_schema
        retries = 0
        while retries < self.MAX_RETRIES:
            try:
                if self.format == "coco":
                    self.load_dataset_from_file()
                else:
                    self.load_data_frame(org)
                    
                spinner.succeed("Data loaded successful!")
                spinner.succeed("Succeed!")
                break  # Exit the loop if initialization succeeds
            except requests.exceptions.RequestException as e:
                print(f"Network error occurred: {str(e)}")
                retries += 1
                if retries < self.MAX_RETRIES:
                    print(f"Retrying in {self.RETRY_DELAY} second(s)...")
                    time.sleep(self.RETRY_DELAY)
                else:
                    spinner.fail("Fail!")
                spinner.stop()
            except PollingException as e:
                spinner.fail(str(e))
                break

    def load_data_frame(self, org=None):
        """
        Loads the data frame, creates a CSV file, zips it, and uploads it to the server.
        """
        self.data_frame_validation()
        create_csv_and_zip_from_data_frame(self.raga_extracted_dataset, self.csv_file, self.zip_file)
        signed_upload_path, file_path = self.get_pre_signed_s3_url(self.zip_file)
        upload_file(
            signed_upload_path,
            self.zip_file,
            success_callback=on_upload_success,
            failure_callback=on_upload_failed,
        )
        delete_files(self.csv_file, self.zip_file)

        self.dataset_file_id = self.create_dataset_load_definition(file_path, "csv", self.raga_schema.columns)
        res_data = self.notify_server(org=org)
        data = res_data.get('data')
        if isinstance(data, dict) and data.get('jobId', None):
            spinner.start()            
            wait_for_status(self.test_session, data.get('jobId', None))            
            if org == "lm-v2":      
                spinner.start()        
                res_data = self.notify_server(org="lm-img-v2")
                data = res_data.get('data')
                if isinstance(data, dict) and data.get('jobId', None):
                    wait_for_status(self.test_session, data.get('jobId', None))
            spinner.stop()

    def load_dataset_from_file(self):
        from raga import REQUIRED_ARG
        if not self.format:
            raise DatasetException(f"{REQUIRED_ARG.format('format')}")
        if not self.model_name:
            raise DatasetException(f"{REQUIRED_ARG.format('model_name')}")
        if not self.inference_col_name:
            raise DatasetException(f"{REQUIRED_ARG.format('inference_col_name')}")
            
        file_dir = os.path.dirname(self.raga_dataset)
        file_name_without_ext, file_extension, file_name = get_file_name(
            self.raga_dataset)
        zip_file_name = os.path.join(file_dir, file_name_without_ext + ".zip")
        with zipfile.ZipFile(zip_file_name, "w") as zip_file:
            zip_file.write(self.raga_dataset, file_name)
        signed_upload_path, file_path  = self.get_pre_signed_s3_url(
            file_name_without_ext + ".zip")
        upload_file(
            signed_upload_path,
            zip_file_name,
            success_callback=on_upload_success,
            failure_callback=on_upload_failed,
        )
        if os.path.exists(zip_file_name):
            os.remove(zip_file_name)
            logger.debug("Zip file deleted")
        else:
            logger.debug("Zip file not found")

        arguments = make_arg(self.model_name, self.inference_col_name, self.embedding_col_name)

        self.dataset_file_id = self.create_dataset_load_definition(file_path, self.format, arguments)
        self.notify_server()
    

    def lightmetrics_data_upload(self, api_version="v1"):
        if api_version == "v1":
            self.load(org="lm")
        else:
            self.load(org="lm-v2")


    def head(self):
        res_data = self.test_session.http_client.post(
            f"api/dataset/data",
            headers={"Authorization": f'Bearer {self.test_session.token}'},
            data={
                "datasetId":self.dataset_id
            }
        )
        if not res_data or 'data' not in res_data or 'rows' not in res_data['data'] or 'columns' not in res_data['data']:
            raise DatasetException("Record not found!")
        
        
        print(self.filter_head(res_data.get('data', {}).get('rows', {}).get('docs', []), res_data.get('data', {}).get('columns', {})))

    def filter_head(self, rows, columns):
        pd_data = pd.DataFrame(rows)
        columns_temp = [col.get("columnName") for col in columns]
        existing_columns = [col for col in columns_temp if col in pd_data.columns]
        return pd_data[existing_columns]

    def get_pre_signed_s3_url(self, file_name: str):
        try:
            res_data = self.test_session.http_client.post(
                f"api/dataset/upload/preSignedUrls",
                {"datasetId": self.dataset_id, "fileNames": [file_name], "contentType": "application/zip"},
                {"Authorization": f'Bearer {self.test_session.token}'},
            )
            if res_data.get("data") and "urls" in res_data["data"] and "filePaths" in res_data["data"]:
                logger.debug("Pre-signed URL generated")
                return res_data["data"]["urls"][0], res_data["data"]["filePaths"][0]
            else:
                error_message = "Failed to get pre-signed URL. Required keys not found in the response."
                logger.error(error_message)
                raise ValueError(error_message)
        except Exception as e:
            logger.exception("An error occurred while getting pre-signed URL: %s", e)
            raise

    def notify_server(self, org=None) -> dict:
        """
        Notifies the server to load the dataset with the provided experiment ID and data definition.
        """
        spinner.start()
        end_point = "api/experiment/load-data"
        if org=="lm":
            end_point = "api/experiment/load-data-lm"
        if org=="lm-v2":
            end_point = "api/experiment/load-data-lm/v2"
        if org=="lm-img-v2":
            end_point = "api/experiment/load-image-data-lm"
            
        res_data = self.test_session.http_client.post(
            end_point,
            {"datasetFileId": self.dataset_file_id},
            {"Authorization": f'Bearer {self.test_session.token}'},
        )
        logger.debug(res_data.get('data', ''))
        return res_data

    

    def create_dataset(self):
        if not self.test_session.project_id:
            raise DatasetException("Project ID is required.")
        if not self.test_session.token:
            raise DatasetException("Token is required.")

        res_data = self.test_session.http_client.post(
            "api/dataset",
            {"name": self.name,
             "projectId": self.test_session.project_id,
             "type": self.type,
             "parentDataset": self.parent_dataset},
            {"Authorization": f'Bearer {self.test_session.token}'},
        )

        if not res_data or 'data' not in res_data or 'id' not in res_data['data']:
            raise DatasetException("Failed to create dataset.")

        return res_data['data']['id']

    def create_dataset_creds(self,):
        if not self.dataset_id:
            raise DatasetException("Dataset ID is required.")
        if not self.test_session.token:
            raise DatasetException("Token is required.")

        data = {
            "datasetId": self.dataset_id,
            "storageService": "s3",
            "json": {'region': self.creds.region}
        }
        res_data = self.test_session.http_client.post(
            "api/dataset/credential",
            data,
            {"Authorization": f'Bearer {self.test_session.token}'},
        )

        if not res_data or 'data' not in res_data or 'id' not in res_data['data']:
            raise DatasetException("Failed to create dataset credentials.")

        return res_data['data']['id']
    
    def create_dataset_load_definition(self, filePath: str, type: str, arguments: dict):
        spinner.start()
        payload = {
            "datasetId": self.dataset_id,
            "filePath": filePath,
            "type": type,
            "arguments": arguments
        }

        res_data = self.test_session.http_client.post(
            "api/dataset/definition", payload,
            {"Authorization": f'Bearer {self.test_session.token}'},
        )
        return res_data.get('data',{}).get('id')

    def get_data_frame(self, columns:list):
        image_id_column = next((item['customerColumnName'] for item in self.raga_schema.columns if item['type'] == 'imageName'), None)
        if image_id_column not in columns:
            columns.append(image_id_column)
        
        missing_columns = [col for col in columns if col not in self.raga_extracted_dataset.columns]
        if not missing_columns:
            return self.raga_extracted_dataset[columns], image_id_column
        else:
            missing_columns_str = ', '.join(missing_columns)
            raise DatasetException(f"The following columns do not exist in the DataFrame: {missing_columns_str}")

    def set_data_frame(self, data_frame:pd.DataFrame):
        image_id_column = next((item['customerColumnName'] for item in self.raga_schema.columns if item['type'] == 'imageName'), None)
        merged_df = pd.merge(self.raga_extracted_dataset, data_frame, on=image_id_column, how='inner', suffixes=('', '_right'))
        self.raga_extracted_dataset = merged_df
        return self.raga_extracted_dataset
    
    def data_frame_validation(self):
        from raga.constants import SCHEMA_KEY_DOES_NOT_EXIST, COL_DOES_NOT_EXIST

        column_list = self.raga_extracted_dataset.columns.to_list()
        for col in self.raga_schema.columns:
            if col.get("customerColumnName") not in column_list:
                raise DatasetException(f"{COL_DOES_NOT_EXIST.format(col.get('customerColumnName'), column_list)}")
            # self.validate_data_frame_value(col, self.raga_extracted_dataset.loc[0,col.get("customerColumnName")])

        if not check_key_value_existence(self.raga_schema.columns, 'type', 'imageName'):
            raise DatasetException(SCHEMA_KEY_DOES_NOT_EXIST)
    
    def validate_data_frame_value(self, col, col_value):
        from raga.constants import DATASET_FORMAT_ERROR

        if col.get('type') == "classification" and "confidence" not in col_value:
                raise DatasetException(f"{DATASET_FORMAT_ERROR.format(col.get('customerColumnName'))}")
        
        if (col.get('type') == "imageEmbedding" or col.get('type') == "roiEmbedding") and "embeddings" not in col_value:
                raise DatasetException(f"{DATASET_FORMAT_ERROR.format(col.get('customerColumnName'))}")
        
        if col.get('type') == "inference" and "detections" not in col_value:
                raise DatasetException(f"{DATASET_FORMAT_ERROR.format(col.get('customerColumnName'))}")
        return 1
    
    def parent_dataset_validation(self):
        if self.parent_dataset:
            res_data = self.test_session.http_client.get(f"api/dataset?projectId={self.test_session.project_id}&name={self.parent_dataset}", headers={"Authorization": f'Bearer {self.test_session.token}'})

            if not isinstance(res_data, dict):
                    raise ValueError(INVALID_RESPONSE)
            dataset_id = res_data.get("data", {}).get("id")

            if not dataset_id:
                raise KeyError(INVALID_RESPONSE_DATA)
            return dataset_id
        return None

    def upload_file_to_s3(self, local_file_path, bucket_name, s3_key, s3):
        parts = s3_key.split('/')
        object_key = '/'.join(parts[-1:])
        try:
            s3.upload_file(local_file_path, bucket_name, s3_key)
        except Exception as e:
            print(f"Cannot upload file to s3, object:{object_key} : {e}")


    def upload_images_to_s3(self, s3, dataframe, column:str, s3_bucket_name:str, s3_key:str):
        for each_row in dataframe[column]:
            local_file_path = each_row
            parts = local_file_path.split('/')
            file_name = '/'.join(parts[-1:])
            s3_file_loc = f'{s3_key}/{file_name}'
            self.upload_file_to_s3(local_file_path, s3_bucket_name, s3_file_loc, s3)

    def __obj_embedding(self, x):
        Embeddings = ImageEmbedding()
        try:
            listEmb = eval(x)
            for emb in listEmb:
                Embeddings.add(Embedding(emb))
            return Embeddings
        except:
            x = '[]'
            listEmb = eval(x)
            for emb in listEmb:
                Embeddings.add(Embedding(emb))
            return Embeddings


    def __obj_mistake_scores(self, x):
        mistake_score = MistakeScore()
        try:
            x = ast.literal_eval(x)
            for key, value in x.items():
                if value.get('area'):
                    mistake_score.add(key=key, value=value.get("mistake_score"), area=value.get('area'))
                else:
                    mistake_score.add(key=key, value=value.get("mistake_score"), area=0)
            return mistake_score
        except:
            x = '{}'
            x = ast.literal_eval(x)
            for key, value in x.items():
                mistake_score.add(key=key, value=value, area=value)
            return mistake_score


    def add_embeddings(self, userDataFrame:pd.DataFrame, model:str, col_name:str):
        df = self.raga_dataset
        schema = self.raga_schema
        if "embedding" not in userDataFrame:
            print("Your dataframe doesn't have embedding column")
            sys.exit(1)
        if col_name in df.columns:
            df=df.drop(columns=[col_name], axis=1)
            new_schema_columns = [column for column in schema.columns if column.get("customerColumnName") != col_name]
            schema.columns = new_schema_columns
        merged_df = pd.DataFrame()
        if "imageId" in df.columns and "imageId" in userDataFrame.columns:
            if df['imageId'].dtype != userDataFrame['imageId'].dtype:
                userDataFrame['imageId'] = userDataFrame['imageId'].astype(df['imageId'].dtype)
            merged_df = pd.merge(df, userDataFrame, on='imageId', how='outer')
        if "imageId" not in df.columns:
            merged_df = pd.concat([df, userDataFrame], axis=1)
            schema.add("imageId", PredictionSchemaElement())
        merged_df[col_name] = merged_df["embedding"].apply(self.__obj_embedding)
        merged_df = merged_df.drop("embedding", axis=1)
        self.raga_dataset = merged_df
        schema.add(col_name, ImageEmbeddingSchemaElement(model=model))
        self.raga_schema = schema
        self.raga_extracted_dataset = data_frame_extractor(self.raga_dataset)


    def add_mistake_scores(self, userDataFrame:pd.DataFrame, col_name:str, refColName:str=""):
        df = self.raga_dataset
        schema = self.raga_schema
        if "mistake_score" not in userDataFrame:
            print("Your dataframe doesn't have mistake_score column")
            sys.exit(1)
        if col_name in df.columns:
            df=df.drop(columns=[col_name], axis=1)
            new_schema_columns = [column for column in schema.columns if column.get("customerColumnName") != col_name]
            schema.columns = new_schema_columns
        merged_df = pd.DataFrame()
        if "imageId" in df.columns and "imageId" in userDataFrame.columns:
            if df['imageId'].dtype != userDataFrame['imageId'].dtype:
                userDataFrame['imageId'] = userDataFrame['imageId'].astype(df['imageId'].dtype)
            merged_df = pd.merge(df, userDataFrame, on='imageId', how='outer')
        if "imageId" not in df.columns:
            merged_df = pd.concat([df, userDataFrame], axis=1)
            schema.add("imageId", PredictionSchemaElement())
        merged_df[col_name] = merged_df["mistake_score"].apply(self.__obj_mistake_scores)
        merged_df = merged_df.drop("mistake_score", axis=1)
        self.raga_dataset = merged_df
        schema.add(col_name, MistakeScoreSchemaElement(ref_col_name=refColName))
        self.raga_schema = schema
        self.raga_extracted_dataset = data_frame_extractor(self.raga_dataset)


class CocoDataset(Dataset):

    def __init__(self, test_session: TestSession, name: str, type: Optional[str],
                 json_file_path: str, image_folder_path:str, s3_bucket_name: str, s3_key: str, aws_raga_access_key: str, aws_raga_secret_key: str,
                 aws_session_token: Optional[str] = None):
        if(type != "object_detection"):
            print("Invalid type entered, please try again with type=object_detection")
            sys.exit(1)
        self.s3_bucket_name = s3_bucket_name
        self.s3_key = s3_key
        self.json_file_path = json_file_path
        self.image_folder_path = image_folder_path
        self.aws_raga_access_key = aws_raga_access_key
        self.aws_raga_secret_key = aws_raga_secret_key
        self.aws_session_token = aws_session_token
        self.inference_file_path = None
        self.s3 = boto3.client('s3', aws_access_key_id=aws_raga_access_key, aws_secret_access_key=aws_raga_secret_key,
                          aws_session_token=aws_session_token if aws_session_token is not None else None)
        pd_data_frame = self.__json_parser(self.json_file_path)
        schema = self.__create_coco_schema()
        super(CocoDataset, self).__init__(test_session=test_session, name=name, type=DATASET_TYPE.IMAGE, data=pd_data_frame, schema=schema)


    def __map_imageId_row_to_annotation(self, grouped_annotations, df, column_name):
        annotations_df_column = []
        for imageId in df["imageId"]:
            annotations_object = ImageDetectionObject()
            if imageId in grouped_annotations:
                i=0
                for detection in grouped_annotations[imageId]:
                    if column_name == "annotations":
                        detection['id'] = str(detection['id'])
                        category_name = self.category_id_to_name_map[detection['category_id']]
                        annotations_object.add(ObjectDetection(Id=detection['id'], ClassId=detection['category_id'], ClassName=category_name, Format="xn,yn_normalised", Confidence=1.0,  BBox= detection['bbox']))
                    else:
                        class_name = self.category_id_to_name_map[int(eval(detection['Class_id']))]
                        annotations_object.add(ObjectDetection(Id=i, ClassId=detection["Class_id"], ClassName=class_name, Confidence=detection['Confidence'], BBox= detection['Bbox'], Format="xn,yn_normalised"))
                        i = i+1
                annotations_df_column.append(annotations_object)
            else:
                annotations_df_column.append(annotations_object)
        df[column_name] = annotations_df_column
        return df[column_name]


    def __get_annotations_by_imageId(self, row):
        grouped_annotations = {}
        annotations = row
        for annotation in annotations:
            image_id = annotation['image_id']
            if image_id not in grouped_annotations:
                grouped_annotations[image_id] = []
            grouped_annotations[image_id].append(annotation)
        return grouped_annotations


    def __create_inference_mapping(self, inferences_file_path):
        import glob
        path = inferences_file_path
        keys = ["Class_id", "x1", "y1", "x2", "y2", "Confidence"]
        row_dict = {}
        for files in glob.glob(path +"/*.txt"):
            filename = os.path.basename(files)
            imageId = os.path.splitext(filename)[0]
            imageId = int(imageId)
            file1 = open(files, 'r')
            lines = file1.readlines()
            row_list = []

            for line in lines:
                values = line.strip().split(',')
                line_dict = dict(zip(keys, values))
                line_dict["Bbox"] = [line_dict["x1"], line_dict["y1"], line_dict["x2"], line_dict["y2"]]
                row_list.append(line_dict)

            row_dict[imageId]=row_list

            file1.close()
        return row_dict

    def __create_category_id_to_name_map(self, category_list):
        self.category_id_to_name_map = {}
        for category_dict in category_list:
            self.category_id_to_name_map[category_dict["id"]] = category_dict["name"]
        return self.category_id_to_name_map

    def __create_s3_url(self):
        return f"https://{self.s3_bucket_name}.s3.amazonaws.com/{self.s3_key}"


    def __make_url(self, imageUri):
        parts = imageUri.split('/')
        imageName = '/'.join(parts[-1:])
        s3_url = self.__create_s3_url()
        return f"{s3_url}/{imageName}"


    def __json_parser(self, json_file):
        with open(json_file, 'r') as f:
            coco_data = json.load(f)
        df = pd.DataFrame()
        df_images = pd.DataFrame(coco_data["images"])
        category_list = coco_data["categories"]
        self.category_id_to_name_map = self.__create_category_id_to_name_map(category_list)
        df["imageId"] = df_images["id"]
        df["imageUri"] = df_images["file_name"].apply(lambda x: f"{self.image_folder_path}/{x}")
        df["width"] = df_images["width"]
        df["height"] = df_images["height"]
        grouped_annotations = self.__get_annotations_by_imageId(coco_data["annotations"])
        df["annotations"] = self.__map_imageId_row_to_annotation(grouped_annotations, df, "annotations")
        return df


    def __create_coco_schema(self):
        schema = RagaSchema()
        schema.add("imageId", PredictionSchemaElement())
        schema.add("imageUri", ImageUriSchemaElement())
        schema.add("width", AttributeSchemaElement())
        schema.add("height", AttributeSchemaElement())
        schema.add("annotations", InferenceSchemaElement(model="gt"))
        return schema


    def __add_inference_schema(self, customer_col_name, model_name):
        schema = self.raga_schema
        new_schema_columns = [column for column in schema.columns if column.get("customerColumnName") != customer_col_name]
        schema.columns = new_schema_columns
        schema.add(customer_col_name, InferenceSchemaElement(model = model_name))
        self.raga_schema = schema


    def __add_inference_column_to_dataframe(self, inferences_file_path, customer_column_name, format):
        df = self.raga_dataset
        if customer_column_name in df.columns:
            df=df.drop(columns=[customer_column_name], axis=1)
        inference_map = self.__create_inference_mapping(inferences_file_path)
        df[customer_column_name] = self.__map_imageId_row_to_annotation(inference_map, df, customer_column_name)
        self.customer_column_name = customer_column_name
        self.raga_dataset = df


    def add_inference(self, inferences_file_path, format, model_name, customer_column_name):
        if str.lower(format) != "yolov5":
            print(f"format {format} is not supported, please try again with format YOLOv5")
            sys.exit(1)
        self.inference_file_path = inferences_file_path
        self.__add_inference_column_to_dataframe(inferences_file_path, customer_column_name, format)
        self.__add_inference_schema(customer_column_name, model_name)

    def upload(self):
        if self.inference_file_path is None:
            print("Uploading images to s3")
            spinner.start()
            self.upload_images_to_s3(self.s3, self.raga_extracted_dataset, "imageUri", self.s3_bucket_name, self.s3_key)
            spinner.stop()
        self.raga_dataset["imageUri"] = self.raga_dataset["imageUri"].apply(lambda x: self.__make_url(x))
        self.raga_extracted_dataset = data_frame_extractor(self.raga_dataset)
        self.load()


class SSDataset(Dataset):
    def __init__(self, test_session: TestSession, name: str, type: Optional[str],
                 annotation_file_path:str, image_folder_path:str, s3_bucket_name: str, images_s3_key: str, annotation_s3_key:str,
                 label_map:dict, aws_raga_access_key: str, aws_raga_secret_key: str,
                 aws_session_token: Optional[str] = None):
        if(type != "semantic_segmentation"):
            print("Invalid type entered, please try again with type=semantic_segmentation")
            sys.exit(1)
        self.s3_bucket_name = s3_bucket_name
        self.images_s3_key = images_s3_key
        self.annotation_s3_key = annotation_s3_key
        self.annotation_file_path = annotation_file_path
        self.image_folder_path = image_folder_path
        self.label_map = label_map
        self.aws_raga_access_key = aws_raga_access_key
        self.aws_raga_secret_key = aws_raga_secret_key
        self.aws_session_token = aws_session_token
        self.inference_file_path = None
        self.inference_s3_key = None
        self.customer_column_name_to_s3_key_map = {}
        self.s3 = boto3.client('s3', aws_access_key_id=aws_raga_access_key, aws_secret_access_key=aws_raga_secret_key,
                               aws_session_token=aws_session_token if aws_session_token is not None else None)
        pd_data_frame = self.__make_df(self.image_folder_path)
        schema = self.__create_semantic_schema()
        super(SSDataset, self).__init__(test_session=test_session, name=name, type=DATASET_TYPE.IMAGE, data=pd_data_frame, schema=schema)


    def __upload_annotations_to_s3(self, column_name, s3_key):
        self.upload_images_to_s3(self.s3, self.raga_extracted_dataset, column_name, self.s3_bucket_name, s3_key)

    def __create_s3_url(self, s3_key):
        return f"https://{self.s3_bucket_name}.s3.amazonaws.com/{s3_key}"


    def __make_s3_url(self, imageName, s3_key):
        s3_url = self.__create_s3_url(s3_key)
        return f"{s3_url}/{imageName}"


    def __make_local_url(self, local_file_path, x):
        local_url = f"{local_file_path}/{x}"
        return local_url


    def __make_df(self, file_path):
        data_frame = pd.DataFrame()
        id_list = []
        for file_name in os.listdir(file_path):
            id_list.append(file_name)
        data_frame["imageId"] = id_list
        data_frame["imageUri"] = data_frame["imageId"].apply(lambda x: self.__make_local_url(self.image_folder_path, x))
        data_frame["annotations"] = data_frame["imageId"].apply(lambda x: self.__make_local_url(self.annotation_file_path, x))
        return data_frame


    def __create_semantic_schema(self):
        schema = RagaSchema()
        schema.add("imageId", PredictionSchemaElement())
        schema.add("imageUri", ImageUriSchemaElement())
        schema.add("annotations", TIFFSchemaElement(label_mapping=self.label_map, schema="tiff"))
        return schema


    def __add_inference_column_to_dataframe(self, customer_column_name):
        df = self.raga_dataset
        if customer_column_name in df.columns:
            df=df.drop(columns=[customer_column_name], axis=1)
        self.customer_column_name = customer_column_name
        df[customer_column_name] = df["imageId"].apply(lambda x: self.__make_local_url(self.inference_file_path, x))
        self.raga_dataset = df
        self.raga_extracted_dataset = data_frame_extractor(self.raga_dataset)


    def __add_inference_schema(self, customer_col_name, model_name):
        schema = self.raga_schema
        new_schema_columns = [column for column in schema.columns if column.get("customerColumnName") != customer_col_name]
        schema.columns = new_schema_columns
        schema.add(customer_col_name, TIFFSchemaElement(label_mapping=self.label_map, schema="tiff", model=model_name))
        self.raga_schema = schema


    def add_inference(self, inferences_file_path, inference_s3_key, model_name, customer_column_name):
        self.inference_file_path = inferences_file_path
        self.inference_s3_key = inference_s3_key
        self.__add_inference_column_to_dataframe(customer_column_name)
        self.customer_column_name_to_s3_key_map[customer_column_name] = inference_s3_key
        self.__add_inference_schema(customer_column_name, model_name)

    def upload(self):
        if self.inference_file_path is None:
            print("Uploading images to s3")
            spinner.start()
            self.upload_images_to_s3(self.s3, self.raga_extracted_dataset, "imageUri", self.s3_bucket_name, self.images_s3_key)
            spinner.stop()
            print("Uploading annotations to s3")
            spinner.start()
            self.__upload_annotations_to_s3("annotations", self.annotation_s3_key)
            spinner.stop()
        else:
            for key, value in self.customer_column_name_to_s3_key_map.items():
                print("Uploading inferences to s3")
                spinner.start()
                self.__upload_annotations_to_s3(key, value)
                spinner.stop()
                self.raga_dataset[key] = self.raga_dataset["imageId"].apply(lambda x: self.__make_s3_url(x, value))
        self.raga_dataset["imageUri"] = self.raga_dataset["imageId"].apply(lambda x: self.__make_s3_url(x, self.images_s3_key))
        self.raga_dataset["annotations"] = self.raga_dataset["imageId"].apply(lambda x: self.__make_s3_url(x, self.annotation_s3_key))
        self.raga_extracted_dataset = data_frame_extractor(self.raga_dataset)
        self.load()



import json
import re

import requests
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult
import pandas as pd
from raga import *
from raga.llm.exporter import constants


class S3Exporter(SpanExporter):
    def __init__(self, test_session: TestSession):
        self.test_session = test_session
        self.trace_spans = dict()
        self.dataframe = pd.DataFrame()

    def export(self, spans):
        for span in spans:
            span_json = json.loads(span.to_json())
            trace_id = span_json.get(constants.CONTEXT).get(constants.TRACE_ID)

            if trace_id not in self.trace_spans:
                self.trace_spans[trace_id] = list()

            self.trace_spans[trace_id].append(span_json)

            if span_json[constants.PARENT_ID] is None:
                trace = self.trace_spans[trace_id]
                trace_uri = self.upload_to_s3(trace_id, trace)
                self.add_to_dataframe(trace_id, trace_uri, trace)
                del self.trace_spans[trace_id]

        return SpanExportResult.SUCCESS

    def add_to_dataframe(self, trace_id, trace_uri, trace):
        data = dict()
        data[constants.TRACE_ID] = trace_id
        data[constants.TRACE_URI] = trace_uri
        for span in trace:
            attributes = span[constants.ATTRIBUTES]
            if constants.OPEN_INFERENCE_SPAN_KIND in attributes and attributes[
                constants.OPEN_INFERENCE_SPAN_KIND] == constants.CHAIN:
                data[constants.PROMPT] = attributes[constants.INPUT_VALUE]
                data[constants.RESPONSE] = attributes[constants.OUTPUT_VALUE]
            elif constants.OPEN_INFERENCE_SPAN_KIND in attributes and attributes[
                constants.OPEN_INFERENCE_SPAN_KIND] == constants.RETRIEVER:
                data[constants.CONTEXT] = attributes[constants.OUTPUT_VALUE]
        trace_dataframe = pd.DataFrame(data, index=[0])
        self.dataframe = pd.concat([self.dataframe, trace_dataframe], ignore_index=True)

    def upload_to_s3(self, trace_id, trace):
        trace_uri, file_path = self.get_pre_signed_s3_url(trace_id)
        json_data = {constants.TRACE: trace}
        response = requests.put(trace_uri, data=json.dumps(json_data), headers={'Content-Type': constants.APPLICATION_JSON})
        if response.status_code == 200:
            print('Upload successful.')
        else:
            print(f'Upload failed with status code: {response.status_code}')
        trace_uri_component_list = trace_uri.split("?")
        if trace_uri_component_list:
            return trace_uri_component_list[0]
        else:
            return None

    def shutdown(self):
        pass

    def get_dataframe(self):
        return self.dataframe

    def get_pre_signed_s3_url(self, file_name: str):
        try:
            res_data = self.test_session.http_client.post(endpoint=f"{constants.GET_PRE_SIGNED_URL_API}",
                                                         data={"projectName": self.test_session.project_name, "fileNames": [f'{file_name}.json'], "contentType": constants.APPLICATION_JSON},
                                                         headers={"Authorization": f'Bearer {self.test_session.token}'}
                                                         )
            data = res_data.get(constants.DATA)
            if data and 'urls' in data and 'filePaths' in data:
                logger.debug("Pre-signed URL generated")
                return data['urls'][0], data['filePaths'][0]
            else:
                error_message = "Failed to get pre-signed URL. Required keys not found in the response."
                logger.error(error_message)
                raise ValueError(error_message)
        except Exception as e:
            logger.exception("An error occurred while getting pre-signed URL for trace id %s: %s", file_name, e)



from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.sdk import trace as trace_sdk
from opentelemetry import trace as trace_api
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace.export import SimpleSpanProcessor
from opentelemetry import trace as trace_api

from openinference.instrumentation.langchain._tracer import OpenInferenceTracer
from wrapt import wrap_function_wrapper
from importlib.util import find_spec
from typing import Type, Callable, Any

from raga.llm.exporter.s3exporter import *
from raga.llm.instrumentation.langchain import constants


class LangChainInstrumentor(BaseInstrumentor):
    def __init__(self, test_session) -> None:
        self.s3_exporter = S3Exporter(test_session)
        if find_spec(constants.LANGCHAIN_CORE) is None:
            raise PackageNotFoundError(
                "Missing `langchain-core`. Install with `pip install langchain-core`."
            )
        super().__init__()

    def instrument(self) -> None:
        tracer_provider = trace_sdk.TracerProvider(
            resource=Resource({"ResourceAttributes.PROJECT_NAME": "get_env_project_name()"})
        )
        tracer_provider.add_span_processor(SimpleSpanProcessor(self.s3_exporter))
        super().instrument(skip_dep_check=True, tracer_provider=tracer_provider)

    def instrumentation_dependencies(self):
        return constants.LANGCHAIN_VERSION

    def _instrument(self, **kwargs) -> None:
        if not (tracer_provider := kwargs.get(constants.TRACER_PROVIDER)):
            tracer_provider = trace_api.get_tracer_provider()
        tracer = trace_api.get_tracer(__name__, constants.INSTRUMENTATOR_VERSION, tracer_provider)

        wrap_function_wrapper(
            module="langchain_core.callbacks",
            name="BaseCallbackManager.__init__",
            wrapper=self.BaseCallbackManagerInit(tracer=tracer, cls=OpenInferenceTracer),
        )

    def _uninstrument(self, **kwargs) -> None:
        pass

    def get_dataframe(self):
        return self.s3_exporter.get_dataframe()

    class BaseCallbackManagerInit:
        __slots__ = ("_tracer", "_cls")

        def __init__(self, tracer: trace_api.Tracer, cls: Type["OpenInferenceTracer"]):
            self._tracer = tracer
            self._cls = cls

        def __call__(
                self,
                wrapped: Callable[..., None],
                instance: "BaseCallbackManager",
                args: Any,
                kwargs: Any,
        ) -> None:
            wrapped(*args, **kwargs)
            for handler in instance.inheritable_handlers:
                if isinstance(handler, self._cls):
                    break
            else:
                instance.add_handler(self._cls(tracer=self._tracer), True)



import logging

from raga import TestSession, spinner
from raga.utils.dataset_util import wait_for_status

logger = logging.getLogger(__name__)

def inference_generator(test_session: TestSession,
                        dataset_name: str,
                        output_type: str,
                        model_name: str,
                        model_inference_col_name: str):
    
    from raga.constants import INVALID_RESPONSE, INVALID_RESPONSE_DATA, REQUIRED_ARG_V2
    
    # validate dataset
    assert isinstance(test_session, TestSession), f"{REQUIRED_ARG_V2.format('test_session', 'instance of the TestSession')}"
    assert isinstance(dataset_name, str) and dataset_name, f"{REQUIRED_ARG_V2.format('dataset_name', 'str')}"
    assert isinstance(output_type, str) and output_type, f"{REQUIRED_ARG_V2.format('output_type', 'str')}"
    assert isinstance(model_name, str) and model_name, f"{REQUIRED_ARG_V2.format('model_name', 'str')}"
    assert isinstance(model_inference_col_name, str) and model_inference_col_name, f"{REQUIRED_ARG_V2.format('model_inference_col_name', 'str')}"

    res_data = test_session.http_client.get(f"api/dataset?projectId={test_session.project_id}&name={dataset_name}", headers={"Authorization": f'Bearer {test_session.token}'})
    if not isinstance(res_data, dict):
            raise ValueError(INVALID_RESPONSE)
    
    dataset_id = res_data.get("data", {}).get("id")
    if not dataset_id:
          raise KeyError(INVALID_RESPONSE_DATA)
    
    payload = {
          "datasetId": dataset_id,
          "outputType": output_type,
          "model": model_name,
          "modelInferenceColName": model_inference_col_name,
    }

    post_url = f"api/experiment/test/inferenceGenerator"
    post_headers = {"Authorization": f'Bearer {test_session.token}'}
    res_data = test_session.http_client.post(post_url, data=payload, headers=post_headers)

    if not isinstance(res_data, dict):
        spinner.stop()
        raise ValueError(INVALID_RESPONSE)
    data = res_data.get('data')
    spinner.start()
    if isinstance(data, dict) and data.get('jobId', None):
        wait_for_status(test_session, data.get('jobId', None))
    spinner.stop()
    spinner.succeed("Succeed!")

    



CONTEXT = "context"
PROMPT = "prompt"
RESPONSE = "response"
TRACE_ID = "trace_id"
PARENT_ID = "parent_id"
TRACE_URI = "trace_uri"
OPEN_INFERENCE_SPAN_KIND = "openinference.span.kind"
ATTRIBUTES = "attributes"
CHAIN = "CHAIN"
INPUT_VALUE = "input.value"
OUTPUT_VALUE = "output.value"
RETRIEVER = "RETRIEVER"
TRACE = "trace"
APPLICATION_JSON = "application/json"
DATA = "data"
UPLOAD_PRE_SIGNED_URL = "uploadPresignedUrl"
FILE_PATH = "filePath"
GET_PRE_SIGNED_URL_API = "/api/project/upload/preSignedUrls/"
